from abc import abstractmethod
from typing import Any, Dict, List, ClassVar, Literal, Type
from logging import getLogger
from pydantic import (
    BaseModel,
    BaseConfig,
    IPvAnyInterface,
    root_validator,
)

LOGGER = getLogger(__name__)


class MyConfig(BaseConfig):
    underscore_attrs_are_private = True
    # extra = Extra.ignore
    # arbitrary_types_allowed = True


class IDInitializableModel(BaseModel):
    _mappings: ClassVar[Dict[Any, Dict[int, Any]]] = {}

    def __init__(self, *args, **kvargs):
        super().__init__(*args, **kvargs)
        self._post_init()

    @root_validator(pre=True)
    def initialize_from_id_if_known(cls, values):
        """Map ID to known vrfs, if already known. Allows for the graphql query to be simplified."""
        if "id" in values:
            id = int(values["id"])
            cls_dict = IDInitializableModel._mappings.get(cls, None)
            if cls_dict is not None:
                existing_instance = cls_dict.get(id, None)
                if existing_instance is not None:
                    return existing_instance.dict()
        return values

    def _register(self, id: int, cls: Type):
        """Register this instance to make it constructible via ID alone."""
        IDInitializableModel._mappings.setdefault(cls, {})

        cls_dict: Dict = IDInitializableModel._mappings.get(cls)  # type: ignore
        cls_dict.setdefault(id, self)

    @abstractmethod
    def _post_init(self):
        pass


class RouteTarget(BaseModel):
    id: int
    name: str

    Config = MyConfig


class VRF(IDInitializableModel):
    id: int
    name: str
    rd: str | None
    import_targets: List[RouteTarget] = []
    export_targets: List[RouteTarget] = []

    def __hash__(self):
        return (self.__class__, self.id).__hash__()

    def _post_init(self):
        self._register(self.id, VRF)

    class Config(MyConfig):
        frozen = True


class Manufacturer(BaseModel):
    id: int
    slug: str

    Config = MyConfig


class DeviceType(BaseModel):
    manufacturer: Manufacturer

    Config = MyConfig


class IPAddressWithMask(BaseModel):
    id: int
    address: IPvAnyInterface
    vrf: VRF | None

    Config = MyConfig


class Interface(IDInitializableModel):
    id: int
    mode: str | None
    name: str
    mgmt_only: bool
    ip_addresses: List[IPAddressWithMask]
    vrf: VRF | None

    def __hash__(self):
        return (self.__class__, self.id).__hash__()

    def _post_init(self):
        self._register(self.id, Interface)

    Config = MyConfig


class LDP(BaseModel):
    id: int
    auto_range: bool
    label_start: int | None
    label_end: int | None
    ldp_enabled_interfaces: List[Interface]


class ASN(BaseModel):
    id: int
    asn: int


class BGP_CE(BaseModel):
    id: int
    ip_addresses: List[IPAddressWithMask]
    asn: ASN


class BGP_Neighbour(BaseModel):
    id: int
    update_source_interface: Interface
    asn: ASN


class AddressFamily(BaseModel):
    id: int
    vpnv4_address_family: bool
    vpnv4_community: Literal['STANDARD', "EXTENDED", "BOTH"] | None
    vpnv6_address_family: bool
    vpnv6_community: Literal['STANDARD', "EXTENDED", "BOTH"] | None

    Config = MyConfig


class BGP_Mesh_Neighbourhood(IDInitializableModel):
    id: int
    bgp_mesh: List[BGP_Neighbour]
    address_family: AddressFamily

    def _post_init(self):
        self._register(self.id, BGP_Mesh_Neighbourhood)

    Config = MyConfig


class BGP_PE(BaseModel):
    id: int
    update_source_interface: Interface
    asn: ASN
    advertised_customers: List[BGP_CE]
    bgp_mesh: BGP_Mesh_Neighbourhood | None


class Device(IDInitializableModel):
    id: int
    name: str | None
    device_type: DeviceType | None
    interfaces: List[Interface]
    ldp: LDP | None
    bgp_pe: BGP_PE | None

    _chosen_management_interface: Interface | None = None

    def _post_init(self):
        self._register(self.id, Device)

    def get_management_interface(self) -> Interface | None:
        if self._chosen_management_interface is None:
            for interface in self.interfaces:
                if interface.mgmt_only:
                    if interface.ip_addresses is not None and len(interface.ip_addresses) > 0:
                        self._chosen_management_interface = interface
                        break
                    else:
                        LOGGER.warning(
                            f"Interface {interface.id} on device {self.id} is a designated management interface, but has no IP associated. Skipped."
                        )
        return self._chosen_management_interface

    Config = MyConfig


class Site(BaseModel):
    id: int
    devices: List[Device]

    Config = MyConfig


class QueryPrerequisites(BaseModel):
    """Data which needs to be loaded into the lookup table ahead of the model of interest"""

    vrf_list: List[VRF]
    interface_list: List[Interface]
    bgp_mesh_list: List[BGP_Mesh_Neighbourhood]



from pathlib import Path
from typing import Iterable
from nornir.core import Nornir
from nornir.core.inventory import Defaults
from nornir.core.task import Task, MultiResult, AggregatedResult
from nornir_napalm.plugins.tasks import napalm_configure

from nac.cisco.protocols import BGPUtil, LDPUtil, VRFUtil
from nac.util.common import BaseNAC, HostMapping, InterfaceConfig

from ..model import *
from .nornir_util import CiscoNornirUtil

LOGGER = getLogger(__name__)


class CiscoNAC(BaseNAC):
    def __init__(self, host_mapping: HostMapping, defaults: Defaults, dry_run: bool) -> None:
        self.defaults = defaults
        self.host_mapping = host_mapping
        self.dry_run = dry_run

    def configure_devices(self, backup_dir: Path | None = None, output_dir: Path | None = None) -> AggregatedResult:
        LOGGER.info(f"Beginning device conifiguration run. Dry run: {self.dry_run}")
        nr: Nornir = CiscoNornirUtil.nornir_setup(hosts=self.host_mapping.get_hosts(), defaults=self.defaults)
        return self._configure_devices(
            nr=nr,
            mapping=self.host_mapping,
            backup_dir=backup_dir,
            output_dir=output_dir,
            dry_run=self.dry_run,
        )

    @staticmethod
    def _configure_device(
        task: Task, mapping: HostMapping, backup_dir: Path | None, output_dir: Path | None, dry_run: bool
    ) -> MultiResult:
        host = task.host
        device = mapping[host]
        host.platform = CiscoNornirUtil.fetch_platform(task)

        if backup_dir:
            result = CiscoNornirUtil.backup_config(task, path=backup_dir)
            result.raise_on_error()

        config = "\n".join(CiscoNAC.generate_config(task=task, device=device))

        if output_dir is not None:
            with open(output_dir / f'{host.name}.conf', 'w+') as f:
                f.write(config)

        # Apply config
        result: MultiResult = task.run(
            napalm_configure,
            dry_run=dry_run,
            replace=False,
            configuration=config,
        )
        return result

    @staticmethod
    def generate_config(task: Task, device: Device) -> List[str]:
        """Generate the commands necessary to set the device into the desired state."""
        # Gather config

        interface_config: InterfaceConfig = InterfaceConfig()

        config: List[str] = []
        config += VRFUtil.generate_config(task, device=device, interface_config=interface_config)
        config += LDPUtil.generate_config(task, device=device, interface_config=interface_config)
        config += BGPUtil.generate_config(task, device=device)

        config += CiscoNAC.generate_interface_config(
            interfaces=device.interfaces,
            interface_config=interface_config,
        )

        return config

    @staticmethod
    def generate_interface_config(interfaces: Iterable[Interface], interface_config: InterfaceConfig) -> List[str]:
        """Generate collected interface config.

        Note: IP address needs to be added after everything else due to VRF changes leading to "no ip address" being added by the router.
        """
        config: List[str] = []
        for i in interfaces:
            if i.name in interface_config.keys():
                config += [f"interface {i.name}"]
                config += interface_config[i.name]
                if i.ip_addresses is not None:
                    for ip in i.ip_addresses:
                        config += [f' ip address {ip.address.ip.exploded} {ip.address.netmask.exploded}']
        return config



from .gen_schema import gen_graphql_schema
from os import environ
from pathlib import Path

schema_file = Path(__file__).parent / "schema.py"

if not schema_file.exists():
    gen_graphql_schema(url=environ.get('GRAPHQL_URL'), token=environ.get('GRAPHQL_TOKEN'), output_file=schema_file)
from .schema import *




from logging import getLogger
from typing import Iterable, Set

from nornir.core.task import Task

from nac.util.common import InterfaceConfig, Util

from ...model import *
from ..nornir_util import CiscoNornirUtil

LOGGER = getLogger(__name__)


class VRFUtil(Util):
    @staticmethod
    def generate_remove_config(task: Task, known_vrfs: Iterable[VRF]) -> List[str]:
        """Generate commands to purge all unknown vrfs from the device."""
        installed_vrfs: List[str] = CiscoNornirUtil.get_configured_vrf_names(task)
        known_vrf_names: List[str] = [vrf.name for vrf in known_vrfs]
        commands: List[str] = []

        for vrf in installed_vrfs:
            if vrf not in known_vrf_names:
                commands += [f"no vrf definition {vrf}"]
        LOGGER.debug(f"Found {len(commands)} vrfs to remove.")
        return commands

    @staticmethod
    def generate_config(task: Task, device: Device, interface_config: InterfaceConfig) -> List[str]:
        # Get known vrfs
        vrfs: Set[VRF] = set()
        config: List[str] = []
        for interface in device.interfaces:
            if interface.vrf is not None:
                vrfs.add(interface.vrf)

        # Remove unknown vrfs
        config += VRFUtil.generate_remove_config(task, known_vrfs=vrfs)

        # Set up VRFs
        for vrf in vrfs:
            # Define new VRF
            config.append(f"vrf definition {vrf.name}")

            if vrf.rd is not None:
                config.append(f" rd {vrf.rd}")

            config += [f" route-target export {t.name}" for t in vrf.export_targets]
            config += [f" route-target import {t.name}" for t in vrf.import_targets]

            # TODO: Source the following settings through netbox somehow
            config += [
                " address-family ipv4",
                " exit-address-family",
                " address-family ipv6",
                " exit-address-family",
            ]

        # Associate interface
        for interface in device.interfaces:
            interface_config.setdefault(interface.name, [])
            if interface.vrf is not None:
                interface_config[interface.name] += [f" vrf forwarding {interface.vrf.name}"]
            else:
                interface_config[interface.name] += [f" no vrf forwarding"]

        return config



from dataclasses import dataclass
from logging import getLogger
from os import environ, makedirs
from pathlib import Path

from nornir.core.inventory import Defaults
from nornir.core.task import AggregatedResult

from nac.cisco import CiscoNAC
from nac.util.common import HostMapping
from nac.util.netbox_api import NetBoxAPI

from .model import *
from .util.graphql_util import GraphQLUtil
from .__version__ import __version__

LOGGER = getLogger(__name__)


@dataclass
class Settings:
    interactive: bool
    output_dir: Path | None
    backup_dir: Path | None
    dry_run: bool

    netbox_url: str
    netbox_token: str
    site_id: int
    ignore_ssl_cert: bool

    ssh_password: str
    ssh_username: str

    def __repr__(self) -> str:
        pass

    def __str__(self) -> str:
        return f"""Settings:
        - netbox-url:             {self.netbox_url}
        - backup:                 {self.backup_dir is not None}
        - dry-run:                {self.dry_run}
        - interactive:            {self.interactive}
        - ignore-ssl-certificate: {self.ignore_ssl_cert}
        - site-id:                {self.site_id}"""


def get_confirmation(prompt: str) -> bool:
    truthy, falsy = False, False
    while not (truthy or falsy):
        proceed = input(f"{prompt} [y/n]: ")
        proceed = proceed.lower().strip()
        truthy = proceed in ('y', 'yes')
        falsy = proceed in ('n', 'no')
    return truthy


def run():
    settings: Settings = parse_cli_arguments()

    NetBoxAPI(settings.netbox_url, token=settings.netbox_token).assert_plugin_installed("argos_netbox", version="0.1")

    LOGGER.info(f"Querying site data for site: {settings.site_id}.")
    site: Site = GraphQLUtil.get_site_data(
        graphql_url=f"{settings.netbox_url}/graphql",
        token=settings.netbox_token,
        site_id=settings.site_id,
        ignore_ssl=settings.ignore_ssl_cert,
    )

    LOGGER.info(f"Data received. Creating host mappings.")

    defaults = Defaults(
        password=settings.ssh_password,
        username=settings.ssh_username,
        platform="ios",
    )

    # Create host mapping
    host_mapping: HostMapping = HostMapping.from_site(site, defaults)
    LOGGER.info(f"Found {len(host_mapping)} devices to configure: {', '.join((str(x) for x in host_mapping.keys()))}")
    LOGGER.info(settings)
    if settings.interactive:
        ok = get_confirmation("Please double-check the above information. Does it appear correct to you?")
        if not ok:
            LOGGER.info("Exiting.")
            exit()

    # Configure devices
    nac = CiscoNAC(
        host_mapping=host_mapping,
        defaults=defaults,
        dry_run=settings.dry_run,
    )
    result: AggregatedResult = nac.configure_devices(
        backup_dir=settings.backup_dir,
        output_dir=settings.output_dir,
    )
    result.raise_on_error()
    LOGGER.info("Operation completed successfully.")


def get_value(env_name: str, interactive: bool, prompt: str = '', critical: bool = False) -> str:
    value = environ.get(env_name, None)
    if value is None:
        if interactive:
            prompt = f'"{env_name}" unset. {prompt} '
            if not critical:
                value = input(prompt)
            else:
                import getpass

                value = getpass.getpass(prompt)
        else:
            raise Exception(f'Environment variable "{env_name}" unset and command is not running as interactive.')
    return value


def parse_cli_arguments() -> Settings:
    '''Parses and handles incoming CLI arguments.'''

    # Import locally to not clutter scope of caller
    import logging
    from argparse import ArgumentParser, SUPPRESS

    # Setup parser
    parser = ArgumentParser(
        prog="argos-deploy",
        description=f"Query NetBox via GraphQL and deploy the documented services. v{__version__}",
        add_help=False,
    )
    parser.add_argument(
        '-h',
        '--help',
        help='Show this help message and exit.',  # Default implementation is not capitalized
        action='help',
        default=SUPPRESS,
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        dest='verbose',
        help='Enable debug output.',
        default=False,
    )
    parser.add_argument(
        '-d',
        '--dry-run',
        action='store_true',
        dest='dry_run',
        help='Prevent making lasting changes to the network devices.',
        default=False,
    )
    parser.add_argument(
        '-f',
        '--file-output',
        dest='output_dir',
        help='Save commands to be applied in specified directory. Can be used with or without "--dry-run". Directory defaults to "./output" if unspecified.',
        nargs='?',
        default=None,
        const='./output',
    )
    parser.add_argument(
        '-b',
        '--backup',
        dest='backup_dir',
        help='Backup current network device configurations to specified directory. Directory defaults to "./backups" if unspecified.',
        nargs='?',
        default=None,
        const='./backups',
    )
    parser.add_argument(
        '-y',
        '--no-interact',
        action='store_false',
        dest='interactive',
        help='Proceed without further interactions, even through warnings.',
        default=True,
    )
    parser.add_argument(
        '--ignore-ssl',
        action='store_true',
        dest='ignore_ssl',
        help='Ignore NetBox SSL certificate check. Only use in safe environment.',
        default=False,
    )

    # Parse
    args, _ = parser.parse_known_args()

    # Apply known settings accordingly
    LOGGER.setLevel(logging.DEBUG if args.verbose else logging.INFO)

    interactive = args.interactive

    site = int(get_value('SITE', interactive, 'Enter NetBox ID of site:'))
    ssh_username = get_value('SSH_USERNAME', interactive, 'Enter SSH username of network devices:')
    ssh_password = get_value('SSH_PASSWORD', interactive, 'Enter SSH password of network devices:', critical=True)

    netbox_url = get_value('NETBOX_URL', interactive, 'Enter NetBox url:')
    netbox_token = get_value('NETBOX_TOKEN', interactive, 'Enter NetBox token:', critical=True)

    ignore_ssl = args.ignore_ssl or (
        environ.get('IGNORE_SSL_CERTIFICATE', 'false').lower() in ('false', '0', 'f', 'n', 'no')
    )

    output_dir = args.output_dir
    if output_dir is not None:
        output_dir = Path(output_dir).absolute()
        makedirs(output_dir, exist_ok=True)  # Create output directory if not exists

    backup_dir = args.backup_dir
    if backup_dir is not None:
        backup_dir = Path(backup_dir).absolute()
        makedirs(backup_dir, exist_ok=True)  # Create output directory if not exists

    # Order of arguments matters.
    return Settings(
        interactive=interactive,
        backup_dir=backup_dir,
        output_dir=output_dir,
        ignore_ssl_cert=ignore_ssl,
        ssh_username=ssh_username,
        ssh_password=ssh_password,
        netbox_url=netbox_url,
        netbox_token=netbox_token,
        site_id=site,
        dry_run=args.dry_run,
    )


if __name__ == "__main__":
    run()



import logging
from .main import run

console_handler = logging.StreamHandler()
formatter = logging.Formatter("[%(levelname)s]: %(message)s")
console_handler.setFormatter(formatter)
console_handler.setLevel(logging.DEBUG)

# Ensure logger logs to console
logging.basicConfig(
    level=logging.INFO,
    handlers=[console_handler],
    force=False,
    format="%(asctime)s (%(name)s) [%(levelname)s]: %(message)s",
)



from pynetbox import api
from pynetbox.core.api import Api

import logging

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.DEBUG)


class NetBoxAPI:
    def __init__(
        self,
        url: str,
        *,
        username: str | None = None,
        password: str | None = None,
        token: str | None = None,
    ):
        self.interface = api(url, token=token)
        if token is None and password is not None:
            self.interface.create_token(username=username, password=password)

    def assert_plugin_installed(self, plugin_name: str, version: str | None = None) -> None:
        """Uses NetBox API to verify if a plugin is installed.

        Provided `version` can use semantic versioning - Eg. `"1"`, `"1.2"` or `"1.2.5"`. Omitted details are ignored.
        """
        plugins = self.interface.plugins.installed_plugins()
        if plugins:
            for pl in plugins:
                if pl.get("package", "") == plugin_name:
                    installed_version = pl.get("version", "")
                    LOGGER.info(
                        f'Plugin "{plugin_name}" version "{installed_version}" detected on remote Netbox instance.'
                    )

                    if version is None:
                        return

                    from semver.version import Version

                    installed = Version.parse(installed_version, optional_minor_and_patch=True)
                    desired = Version.parse(version, optional_minor_and_patch=True)

                    if (
                        desired.major == installed.major
                        and (desired.minor == 0 or desired.minor == installed.minor)
                        and (desired.patch == 0 or desired.patch == installed.patch)
                    ):
                        return

                    raise Exception(
                        f'Plugin "{plugin_name}" expected to have version {desired}, but was {installed}. Exiting.'
                    )
        raise Exception("Plugin is not among installed plugins in Netbox instance. Exiting.")



import logging
import re
from pathlib import Path
from re import Pattern
from typing import Dict, List

from nornir.core.inventory import Host
from nornir.core.task import MultiResult, Task
from nornir_napalm.plugins.tasks import napalm_get
from nornir_utils.plugins.tasks.files import write_file

from nac.util.common import BaseNornirUtil

LOGGER = logging.getLogger(__name__)


class CiscoNornirUtil(BaseNornirUtil):
    _config_map: Dict[Host, str] = {}
    _facts_map: Dict[Host, Dict[str, str]] = {}

    @staticmethod
    def get_device_config(task: Task) -> str:
        config = CiscoNornirUtil._config_map.get(task.host, None)
        if config is None:
            # Get current config
            result = task.run(task=napalm_get, getters=["config"])
            result.raise_on_error()
            config = result.result["config"]["running"]
            CiscoNornirUtil._config_map.setdefault(task.host, config)
        return config

    @staticmethod
    def get_device_facts(task: Task) -> Dict[str, str]:
        facts: Dict[str, str] = CiscoNornirUtil._facts_map.get(task.host, None)  # type: ignore
        if facts is None:
            result: MultiResult = task.run(task=napalm_get, getters=["facts"])
            result.raise_on_error()
            facts = result.result["facts"]
            CiscoNornirUtil._facts_map.setdefault(task.host, facts)
        return facts

    @staticmethod
    def backup_config(task: Task, path: str | Path) -> MultiResult:
        LOGGER.info(f'Taking backup of {task.host.name}')
        config = CiscoNornirUtil.get_device_config(task)

        Path(path).mkdir(parents=True, exist_ok=True)

        return task.run(task=write_file, content=config, filename=f"{path}/{task.host.name}.log")

    @staticmethod
    def _filter_config(task: Task, pattern: Pattern) -> List[str]:
        config = CiscoNornirUtil.get_device_config(task)
        return BaseNornirUtil._filter_for_line(config, pattern)

    @staticmethod
    def get_configured_vrf_names(task: Task) -> List[str]:
        return CiscoNornirUtil._filter_config(task, re.compile(pattern=r"^vrf definition (.+)"))

    @staticmethod
    def get_configured_bgp_asns(task: Task) -> List[str]:
        return CiscoNornirUtil._filter_config(task, re.compile(pattern=r"^router bgp (\d+)"))

    @staticmethod
    def fetch_platform(task: Task) -> str:
        facts: Dict[str, str] = CiscoNornirUtil.get_device_facts(task)
        vendor: str = facts["vendor"]
        vendor_map = {
            "cisco": "ios",
            "juniper": "junos",
            "palo_alto_networks": "panos",
        }
        return vendor_map[vendor.lower()]



import ssl
from pathlib import Path
from typing import Any
from urllib.request import urlopen

from sgqlc.endpoint.http import HTTPEndpoint

from nac.constants import ROOT_DIR

from .jinja_util import JinjaUtil
from ..model import *


class GraphQLUtil:
    @staticmethod
    def get_site_data(graphql_url: str, token: str, site_id: int, ignore_ssl: bool = False) -> Site:
        data = GraphQLUtil._send_request(graphql_url, token, site_id, ignore_ssl)
        return GraphQLUtil._parse_query_data(data, site_id)

    @staticmethod
    def _parse_query_data(data: dict, site_id: int) -> Site:
        # Gather & init prerequisites
        vrf_list = GraphQLUtil._get_or_raise(data, "vrf_list")
        interface_list = GraphQLUtil._get_or_raise(data, "interface_list")
        bgp_mesh_list = GraphQLUtil._get_or_raise(data, "bgp_mesh_list")
        QueryPrerequisites(
            vrf_list=vrf_list,
            interface_list=interface_list,
            bgp_mesh_list=bgp_mesh_list,
        )  # Initializing is enough to init datastructure map

        # Gather site data
        site_data = data.get("site", None)
        if site_data is None:
            raise Exception(f'Site with ID "{site_id}" does not appear to exist.')
        return Site(**site_data)

    @staticmethod
    def _get_or_raise(data: dict, field_name: str) -> Any:
        field = data.get(field_name, None)
        if field is None:
            raise Exception(f'Expected "{field_name}" element.')
        return field

    @staticmethod
    def _send_request(graphql_url: str, token: str, site_id: int, ignore_ssl: bool) -> dict:
        query_string = JinjaUtil.render_jinja_template(
            root_file=ROOT_DIR / "queries" / "main.gql",
            site_id=site_id,
        )
        headers = {"Authorization": f"Token {token}", "Accept": "application/json"}

        endpoint = HTTPEndpoint(
            graphql_url,
            headers,
            urlopen=GraphQLUtil.__urlopen_without_ssl_check if ignore_ssl else None,
            method="GET",
        )
        response = endpoint(query=query_string)
        errors = response.get("errors", None)
        if errors is not None:
            raise Exception("Errors in grapql query:\n" + "\n".join((e["message"] for e in errors)))
        return response.get("data", {})

    @staticmethod
    def __urlopen_without_ssl_check(*args, **kvargs) -> Any:
        '''
        Used to onkeypatch ssl certificate verification in case of locally running NetBox instance.
        '''
        ssl_context = ssl.create_default_context()
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE

        return urlopen(*args, context=ssl_context, **kvargs)  # type: ignore



from dataclasses import dataclass
from logging import getLogger

from nornir.core.task import Task

from nac.util.common import Util

from ...model import *
from ..nornir_util import CiscoNornirUtil

LOGGER = getLogger(__name__)


class BGPUtil(Util):
    @staticmethod
    def generate_remove_config(task: Task, known_asn: ASN | None) -> List[str]:
        """Generate commands to purge previously set config from the device."""
        installed_bgp_asns: List[str] = CiscoNornirUtil.get_configured_bgp_asns(task)
        commands: List[str] = []

        for asn in installed_bgp_asns:
            if known_asn is None or int(asn) != known_asn.asn:
                commands.append(f"no router bgp {asn}")
        LOGGER.debug(f"Found {len(commands)} bgps to remove.")
        return commands

    @staticmethod
    def generate_config(task: Task, device: Device) -> List[str]:
        if device.bgp_pe is None:
            return BGPUtil.generate_remove_config(task, known_asn=None)

        bgp_pe: BGP_PE = device.bgp_pe

        ret: List[str] = []
        ret += BGPUtil.generate_remove_config(task, known_asn=bgp_pe.asn)
        ret += [
            f"router bgp {bgp_pe.asn.asn}",
            " no bgp default ipv4-unicast",
            " address-family ipv4",
            " exit-address-family",
            " address-family ipv6",
            " exit-address-family",
        ]
        ret += BGPUtil.generate_vrf_subconfig(task, bgp_pe)
        ret += BGPUtil.generate_bgp_neighbour_subconfig(device)

        return ret

    @staticmethod
    def generate_bgp_neighbour_subconfig(device: Device) -> List[str]:
        """
        Precondition: device has a valid, not-None BGP_PE member field.
        """
        config: List[str] = []
        vpnv4_config: List[str] = []
        vpnv6_config: List[str] = []

        bgp_pe: BGP_PE = device.bgp_pe  # type: ignore
        mesh = bgp_pe.bgp_mesh
        if mesh is None:
            raise Exception(f'BGP PE instance of "{device.name}" does not appear to be part of any BGP mesh.')

        @dataclass
        class ValidatedNeighbour:
            address: IPvAnyInterface
            asn: ASN

        valid_neighbours: List[ValidatedNeighbour] = []

        for neighbour in mesh.bgp_mesh:
            if neighbour.id != bgp_pe.id:
                if neighbour.update_source_interface.ip_addresses is None:
                    raise Exception(
                        f'Expected "ip_addresses" list in BGP PE neighour with ID "{neighbour.id}" in mesh with ID "{mesh.id}".'
                    )
                for address in neighbour.update_source_interface.ip_addresses:
                    valid_neighbours.append(ValidatedNeighbour(address=address.address, asn=neighbour.asn))

        for neighbour in valid_neighbours:
            config += [
                f' neighbor {neighbour.address.ip} remote-as {neighbour.asn.asn}',
                f' neighbor {neighbour.address.ip} update-source {bgp_pe.update_source_interface.name}',
            ]
            if mesh.address_family.vpnv4_address_family == True:
                community = mesh.address_family.vpnv4_community
                if community is None:
                    community = ''
                vpnv4_config += [
                    f"  neighbor {neighbour.address.ip} activate",
                    f"  neighbor {neighbour.address.ip} send-community {community.lower()}",
                ]
            if mesh.address_family.vpnv6_address_family == True:
                community = mesh.address_family.vpnv4_community
                if community is None:
                    community = ''
                vpnv6_config += [
                    f"  neighbor {neighbour.address.ip} activate",
                    f"  neighbor {neighbour.address.ip} send-community {community.lower()}",
                ]

        if mesh.address_family.vpnv4_address_family == True:
            config += [" address-family vpnv4"]
            config += vpnv4_config
            config += [" exit-address-family"]
        if mesh.address_family.vpnv6_address_family == True:
            config += [" address-family vpnv6"]
            config += vpnv6_config
            config += [" exit-address-family"]

        return config

    @staticmethod
    def generate_vrf_subconfig(task: Task, bgp_pe: BGP_PE) -> List[str]:
        ret: List[str] = []

        @dataclass
        class Customer:
            address: str
            asn: int

        mapping: Dict[str, Dict[int, List[Customer]]] = {}
        """{<vrf_name>:{ip_v:[<customers>]}}"""

        # Generate map to avoid customers in same vrf overwriting eachother in the config.

        for advertised_customer in bgp_pe.advertised_customers:
            assert advertised_customer.asn is not None
            for address in advertised_customer.ip_addresses:
                if address.vrf is not None:
                    vrf_name: str = address.vrf.name
                    version: int = address.address.version
                    mapping.setdefault(vrf_name, {})
                    mapping[vrf_name].setdefault(version, [])
                    mapping[vrf_name][version].append(
                        Customer(address=address.address.ip, asn=advertised_customer.asn.asn)
                    )

        for vrf_name, x in mapping.items():
            for ip_v, customers in x.items():
                ret += [f" address-family ipv{ip_v} vrf {vrf_name}"]
                for c in customers:
                    ret += [
                        f"  neighbor {c.address} remote-as {c.asn}",
                        f"  neighbor {c.address} activate",
                    ]
                ret += [" exit-address-family"]
        return ret



from logging import getLogger

from nornir.core.task import Task

from nac.util.common import InterfaceConfig, Util

from ...model import *
from ..nornir_util import CiscoNornirUtil

LOGGER = getLogger(__name__)


class LDPUtil(Util):
    @staticmethod
    def generate_remove_config(task: Task) -> List[str]:
        """Generate commands to purge previously set config from the device."""
        return ["no mpls label range"]

    @staticmethod
    def generate_config(task: Task, device: Device, interface_config: InterfaceConfig) -> List[str]:
        if device.ldp is None:
            return LDPUtil.generate_remove_config(task)
        else:
            ret = []
            if device.ldp.auto_range:
                ret += ["no mpls label range"]  # removes any hard-coded ranges, resulting in auto-range
            else:
                if device.ldp.label_start is None or device.ldp.label_end is None:
                    raise Exception(
                        f"Expected a range to be provided. Got: [{device.ldp.label_start};{device.ldp.label_end}]"
                    )
                ret += [f"mpls label range {device.ldp.label_start} {device.ldp.label_end}"]
            ret += LDPUtil.generate_enable_mpls_on_interface_config(device, interface_config)
            return ret

    @staticmethod
    def generate_enable_mpls_on_interface_config(
        device: Device,
        interface_config: InterfaceConfig,
    ) -> List[str]:
        ret: List[str] = []
        if device.ldp is not None:
            for interface in device.interfaces:
                interface_config.setdefault(interface.name, [])
                if interface in device.ldp.ldp_enabled_interfaces:
                    interface_config[interface.name] += [" mpls ip"]
                else:
                    interface_config[interface.name] += [" no mpls ip"]

        return ret



from sgqlc.endpoint.http import HTTPEndpoint
from sgqlc.codegen.schema import CodeGen
from pathlib import Path


def gen_graphql_schema(url: str, token: str, output_file: str | Path):
    headers = {'Authorization': f'Token {token}', 'Accept': 'application/json'}
    with open(Path(__file__).parent / "schema_query.gql") as f:
        query = f.read().replace('\n', '')

    endpoint = HTTPEndpoint(url, headers, method='GET')
    res = endpoint(query=query)

    def filter_response_dict(schema):
        if not isinstance(schema, dict):
            raise SystemExit('schema must be a JSON object')

        if schema.get('types'):
            return schema
        elif schema.get('data', {}).get('__schema', None):
            return schema['data']['__schema']  # plain HTTP endpoint result
        elif schema.get('__schema'):
            return schema['__schema']  # introspection field
        else:
            raise SystemExit('schema must be introspection object or query result')

    with open(output_file,'w+') as f:
        a = CodeGen(schema=filter_response_dict(res), schema_name='schema', writer=f.write, docstrings=False)
        a.write()



import json
from pathlib import Path

from nornir.core.inventory import Defaults
from nornir.core.task import Task
from pytest import MonkeyPatch

from nac.cisco import CiscoNAC, CiscoNornirUtil
from nac.model import *
from nac.util.common import HostMapping
from nac.util.graphql_util import GraphQLUtil

with open(Path(__file__).parent / "site_sample_data.json") as f:
    data = json.load(f)["data"]
    site: Site = GraphQLUtil._parse_query_data(data, site_id=1)
    mapping: HostMapping = HostMapping.from_site(site)


def get_device_config(task: Task):
    with open(Path(__file__).parent / "configs" / f"{task.host.name}.log") as f:
        return f.read()


def get_expected_config(task: Task) -> List[str]:
    with open(Path(__file__).parent / "configs" / f"{task.host.name}_expected.log") as f:
        return [s for s in f.read().split("\n") if s != "" and not s.startswith('!')]


def test_vrf_config_generation(monkeypatch: MonkeyPatch):
    ran_to_completion = 0

    def test(task: Task, mapping: HostMapping, **kvargs):
        nonlocal ran_to_completion
        monkeypatch.setitem(CiscoNornirUtil._config_map, task.host, get_device_config(task))
        device: Device = mapping[task.host]
        config: List[str] = CiscoNAC.generate_config(task=task, device=device)
        expected = get_expected_config(task)

        # Compare against expected config, disregarding order
        # TODO: is instantiation order not deterministic or why can the order change run to run?
        sorted_config = config.copy()
        sorted_config.sort()
        sorted_expected = expected.copy()
        sorted_expected.sort()
        for cmd, exp in zip(sorted_config, sorted_expected, strict=True):
            if not cmd == exp:
                print("\n")
                print(f"{'#'*20} EXPECTED {'#'*20}")
                print("\n".join(expected))
                print(f"{'#'*20} GOT {'#'*20}")
                print("\n".join(config))
                assert cmd == exp
        ran_to_completion += 1

    nac = CiscoNAC(host_mapping=mapping, defaults=Defaults(), dry_run=True)
    monkeypatch.setattr(nac, "_configure_device", test)
    nac.configure_devices()
    assert ran_to_completion == len(mapping), "One or more configuration dryruns failed to execute fully."



from pathlib import Path

from jinja2 import Template


class JinjaUtil:
    @staticmethod
    def render_jinja_template(root_file: str | Path, **kvargs) -> str:
        with open(Path(root_file)) as f:
            query_source = f.read()
        query: Template = Template(source=query_source)
        query_string = query.render(**kvargs)
        return query_string



from abc import ABC, abstractmethod
from logging import getLogger
from pathlib import Path
from re import Pattern
from typing import List

from nornir.core import Nornir
from nornir.core.inventory import Defaults, Host, Hosts, Inventory
from nornir.core.task import AggregatedResult, Task

from nac.model import Device, Site

LOGGER = getLogger(__name__)


class HostMapping(dict[Host, Device]):
    @staticmethod
    def from_site(site: Site, nornir_defaults: Defaults = Defaults()) -> 'HostMapping':
        mapping = HostMapping()
        for device in site.devices:
            management_interface = device.get_management_interface()
            if management_interface is not None:
                name = device.name if device.name else f"_device_{device.id}"
                host = Host(
                    name=name,
                    hostname=management_interface.ip_addresses[0].address.ip.exploded,  # type: ignore
                    defaults=nornir_defaults,
                )
                mapping[host] = device
            else:
                LOGGER.warning(
                    f'Device (id: {device.id}, name: "{device.name}") does not have a valid "Management Only" interface with associated IP address. Skipping its configuration.'
                )
        return mapping

    def get_hosts(self) -> Hosts:
        hosts = dict()
        for host in self.keys():
            hosts[host.name] = host
        return Hosts(hosts)


class InterfaceConfig(dict[str, list[str]]):
    '''Interface specific configs.'''


class Util(ABC):
    @staticmethod
    @abstractmethod
    def generate_config(task: Task, device: Device) -> List[str]:
        """Generate commands to achieve the desired device state for this protocol."""


class BaseNornirUtil:
    __registered: bool = False

    @staticmethod
    def __register_napalm_plugin() -> None:
        if not BaseNornirUtil.__registered:
            from nornir.core.plugins.connections import ConnectionPluginRegister
            from nornir_napalm.plugins.connections import CONNECTION_NAME, Napalm

            ConnectionPluginRegister.register(name=CONNECTION_NAME, plugin=Napalm)

    @staticmethod
    def nornir_setup(hosts: Hosts, defaults: Defaults = Defaults()) -> Nornir:
        from nornir.plugins.runners import SerialRunner

        BaseNornirUtil.__register_napalm_plugin()

        inv = Inventory(hosts=hosts, defaults=defaults)
        nornir = Nornir(inventory=inv, runner=SerialRunner())  # type: ignore
        return nornir

    @staticmethod
    def _filter_for_line(string: str, pattern: Pattern) -> List[str]:
        matches: List[str] = []
        for line in string.splitlines():
            match = pattern.match(line)
            if match is not None:
                matches.append(*match.groups())
        return matches


class BaseNAC(ABC):
    @staticmethod
    @abstractmethod
    def _configure_device(
        task: Task,
        mapping: HostMapping,
        backup_dir: Path | None,
        output_dir: Path | None,
        **kvargs,
    ) -> AggregatedResult:
        """Generate the commands necessary to set the device into the desired state.

        This function is called automatically for every host."""

    @staticmethod
    @abstractmethod
    def generate_config(task: Task, device: Device) -> List[str]:
        """Generate the commands necessary to set the device into the desired state."""

    def _configure_devices(
        self,
        *,
        nr: Nornir,
        **kvargs,
    ) -> AggregatedResult:
        '''Calls `_configure_device` for each host with the provide kvargs.'''
        try:
            # For each device, configure it
            return nr.run(task=self._configure_device, **kvargs)
        finally:
            nr.close_connections()
